You can search for identifiers within the package.
in-package search v0.2.0
par_incr
Par_incr.Debug
val attach : fn:('a -> string) -> 'a t -> 'a t
Attach to_string function to an incremental. This is especially useful when we want to print the computation tree with dump_tree. With a specialized to_string attached to an incremental, we can get better tree dumps.
to_string
dump_tree
incremental